;;; -*- auto-recompile: t -*-

;; This file is not part of GNU Emacs.
;; 
;; squeak-interaction.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; squeak-interaction.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;-------------
;;             Surmulot integration 

;; last modified February 9, 2008
;; for Emacs 22
;;; --------------------------------------------------------------------------------

(require 'squeak)    
(require 'csound-x)

(defun muo-connected-p ()
  (squeak-connected-somehow-p))

; (muo-get "'a,e,f' kmusic asBox")
(defun muo-get (object)
  (squeak-do (format "SurmulotManager emacsSend: (%s)" object)))

(defun muo-edit-string (encoded-string rid mode label)
  "Decode ENCODED-STRING and create a buffer to edit it
RID is a numeric handle allowing O to know what to do with the edited string
MODE is the major mode the edit buffer should be in
LABEL is the name for that buffer

This function is required by EmacsInteraction in O"
  (save-excursion
    (switch-to-buffer-other-frame
     (generate-new-buffer label))
    (erase-buffer)
    (insert (squeak-decode encoded-string))
    (goto-char (point-min))
    (condition-case nil
	(funcall mode)
      (error nil))
    (set (make-local-variable 'receiver-id) rid)
    (setq header-line-format
	  (list
	   (squeak-console-button-do 
	    " Accept " "accept the changes"
	    (squeak-do
	     (format "EmacsInteraction updateString: '%s' receiverId: %s " 
		     (squeak-encode-buffer)
		     receiver-id))
	    (set-buffer-modified-p nil))
	   " "
	   (squeak-console-button-do 
	    " Explore receiver " "open an ObjectExplorer in Squeak"
	    (squeak-do 
	     (format "(EmacsInteraction msDict at: %s) explore" receiver-id)))
	   " "
	   (squeak-console-button-do 
	    " Accept and exit " "accept the changes and kill this buffer"
	    (squeak-do 
             (format "EmacsInteraction updateString: '%s' receiverId: %s cleanUp: true" 
                     (squeak-encode-buffer)
		     receiver-id))
	    (let ((f (selected-frame)))
	      (when (and (kill-buffer (current-buffer))
			 (eq f (selected-frame)))
		(delete-frame))))
	   " "
	   (squeak-console-button-do 
	    " Exit " "kill this buffer and its frame"
	    (squeak-do 
	     (format "EmacsInteraction cleanUpReceiverId: %s" receiver-id))
	    (let ((f (selected-frame)))
	      (when (and (kill-buffer (current-buffer))
			 (eq f (selected-frame)))
		(delete-frame))))))
    (set-buffer-modified-p nil)))

;; (defun muo-get-buffer ()
;;    "send to O the appropriate MGraphElement for representing the current buffer"
;;    (interactive)
   
;; )

(defun muo-get-sco ()
  "send to O the appropriate MGraphElement for representing the current score"
  (interactive)
  (let ((filename (buffer-file-name)))
    (if (and filename (string-match "\\.sco$" filename))
        (muo-get (format "MGraphMusicFile newFromFile: '%s'" (squeak-filename filename)))
      (muo-get (format "MGraphCsoundScore newFromB64: '%s'" 
                          (if (cscsd-buffer-is-a-sco-p)
                              (squeak-encode-region (cscsd-sco-beginning) (cscsd-sco-end))
                            (squeak-encode-buffer)))))))

(defun muo-get-orc ()
  (interactive)
  "send to O the appropriate MGraphElement for representing the current orchestra"
  (let ((filename (buffer-file-name)))
    (if (and filename (string-match "\\.orc$" filename))
        (muo-get (format "MGraphCsoundComposer newFromFile: '%s'" (squeak-filename filename)))
      (muo-get (format "MGraphCsoundComposer newFromB64: '%s'" 
                          (if (cscsd-buffer-is-an-orc-p)
                              (squeak-encode-region (cscsd-orc-beginning) (cscsd-orc-end))
                            (squeak-encode-buffer)))))))

(defun muo-get-csd ()
  (interactive)
  "send to O the appropriate MGraphElement for representing the current csd"
  (let ((filename (buffer-file-name)))
    (if (and filename (string-match "\\.csd$" filename))
        (muo-get (format "MGraphMusicFile newFromFile: '%s'" (squeak-filename filename)))
      (muo-get (format "MGraphCsoundComposition newFromB64: '%s'" 
			  (squeak-encode-buffer))))))


;;; ============================ testing

(defun surmulot-assert (test result)
  (let ((val (squeak-eval test)))
    (or (string= val result) val)))	

(defun surmulot-tests ()
  (and
   (surmulot-assert "1+1" "2")
   (surmulot-assert "SurmulotManager askEmacsFor: '(+ 8 7)'" "15")
   ;; hairy... and fails !!
;   (surmulot-assert 
;    "SurmulotManager askEmacsFor: '(surmulot-assert \"2+2\" \"4\")'" "t")
))

;;; ============================ meta: surmulot system description

(defvar this-surmulot-version nil
  "holds the current surmulot version number")

(defvar this-surmulot-elisp-version nil
  "holds the current stef-elisp version number")

(defvar this-surmulot-muo-version nil
  "holds the current O version number")

(defvar this-surmulot-gm-version nil
  "holds the current GeoMaestro version number")

(defvar this-surmulot-published-files nil
  "holds the list of all files in the current surmulot release")

(defun surmulot-fetch-meta-info ()
  (when (boundp 'surmulot-directory) ;; this is defined in surmulot.emacs
    (load (expand-file-name "meta/surmulot-org.el" surmulot-directory) t)))

(surmulot-fetch-meta-info)

(defun surmulot-check-up ()
  (let ((modified))
    (dolist (fspec this-surmulot-published-files)
      (when (consp fspec)
        (let* ((file (car fspec))
               (fname (expand-file-name file surmulot-directory))
               (fmd5 (cdr fspec)))
          (unless (and (file-exists-p fname)
                       (string= fmd5 (with-temp-buffer
                                       (insert-file-contents fname)
                                       (md5 (current-buffer) nil nil 'raw-text))))
            (add-to-list 'modified file)))))
    (or modified t)))

;;TO DO:
;;       repair from archive (use 7za) !!


;;; ============================ widgets: low-level functions

(defstruct surmulot-widget name port methods) ;....


(defcustom surmulot-squeak-vm "$SURMULOTDIR/squeak/vm/squeak.exe"
  "Squeak VM for Surmulot"
  :type 'string
  :group 'spfa-paths
  :group 'squeak)

(defun surmulot-squeak-vm ()
  (expand-file-name (substitute-in-file-name surmulot-squeak-vm)))

(defcustom surmulot-squeak-widget-image "$SURMULOTDIR/squeak/muo.image"
  "Squeak image to start when requesting a widget"
  :type 'string
  :group 'spfa-paths
  :group 'squeak)

(defun surmulot-squeak-widget-image ()
  (expand-file-name (substitute-in-file-name surmulot-squeak-widget-image)))

(defcustom surmulot-squeak-reserved-ports-start 6000
  "TCP/IP local port numbers for widgets start at this value upward"
  :type 'integer
  :group 'squeak)

(defun surmulot-make-widget (command &optional width height)
  "have Squeak evaluate COMMAND, which should return a Morph,
and return a surmulot-widget
use function `surmulot-widget-interface' to communicate with the widget"
  (let ((port (surmulot-new-squeak-widget-port))) 
    (save-window-excursion
      (with-temp-buffer ;;  cause du cd !
        (cd (file-name-directory (surmulot-squeak-vm)))
        (shell-command
         (format "%s %s free doIt \"%s\" &"
                 (shell-quote-argument (surmulot-squeak-vm))
                 (shell-quote-argument (surmulot-squeak-widget-image))
                 (format "MuODeployer makeWidget: (%s) extent: (%s@%s) port: %s" 
                         command
                         (or width 500) (or height 300)
                         port))
         (generate-new-buffer-name " *surmulot widget*"))))
    (dotimes (n 30) 
      (sit-for 1)
      (condition-case nil
          (squeak-tcpip-connect port)
        (error nil)))
    (squeak-tcpip-connect port)
    (let ((widget (make-surmulot-widget :port port)))
      (setf (surmulot-widget-methods widget)
            (squeak-eval-with-return "Widget surmulotWidgetMethods" port))
      widget)))

(defun surmulot-new-squeak-widget-port ()
  (loop for port = surmulot-squeak-reserved-ports-start then (1+ port)
        for used in (sort (surmulot-squeak-active-ports) '<)
        if (> used port) return port
        finally return (1+ port)))

(defun surmulot-squeak-active-ports ()
  (save-excursion
    (loop for b in (buffer-list)
          do (set-buffer b)
          if (processp squeak-tcpip-process)
          collect (squeak-port-for-process squeak-tcpip-process))))

;; (surmulot-make-widget "MuO EnvelopeEditor")

;;; ============================ widgets: high-level functions

(defvar surmulot-widgets-list '()
  "All currently opened widgets")

(defvar surmulot-widget-specs 
  '(("Envelope Editor" "EnvelopeEditor new"
     ("insert as f-table" 
      (lambda (widget)
	`(insert (surmulot-widget-do ,widget "Widget asCsoundTable"))))
     ("insert as linseg" 
      (lambda (widget) 
	`(insert (surmulot-widget-do ,widget "Widget envelope asLinseg")))))
    )
  "Different widgets, how to invoke them, how to use them")

(defun surmulot-widgets-add (name command)
  (let ((widget (surmulot-make-widget command)))
    (add-to-list 'surmulot-widgets-list widget)
    (setf (surmulot-widget-name widget) name)))

(defun surmulot-make-widgets-menu ()
  (easy-menu-define surmulot-widgets-menu global-map
    "Squeak widgets"
    `("Widgets"
      ("New widget"
       ,@(mapcar 
          (lambda (spec) `[,(car spec) (surmulot-widgets-add ,(car spec) ,(cadr spec)) t])
          surmulot-widget-specs))
      "--"
      ,@(mapcar 
         (lambda (widget)
           `(,(format (concat (surmulot-widget-name widget) " (port %d)") (surmulot-widget-port widget))
             ["close" (surmulot-widget-close ,widget) t]
             ,@(mapcar
                (lambda (spec)
                  `[,(car spec) ,(funcall (cadr spec) widget) t])
                (cddr (assoc (surmulot-widget-name widget) surmulot-widget-specs)))))
         surmulot-widgets-list))))

(defun surmulot-widget-close (widget)
  (setq surmulot-widgets-list (delete widget surmulot-widgets-list))
  (squeak-eval-no-return "Smalltalk quitPrimitive" (surmulot-widget-port widget)))

(defun surmulot-widget-do (widget command)
   (squeak-eval-with-return command (surmulot-widget-port widget)))

(surmulot-make-widgets-menu)
(add-hook 'activate-menubar-hook 'surmulot-make-widgets-menu)


;;; ==============================================================

(defun surmulot-use-system-csound ()
  "experimental..."
  (interactive)
  (custom-set-default 'cscsd-csound-binary "csound")
  (cscsd-open-to-environment))

;;; ==============================================================


(defcustom surmulot-timidity-cfg "$SURMULOTDIR/timidity/timidity.cfg"
  "Timidity configuration file for Surmulot"
  :type 'file
  :group 'spfa-paths)

(defun surmulot-timidity-cfg ()
  (expand-file-name (substitute-in-file-name surmulot-timidity-cfg)))

(defcustom surmulot-timidity-man "$SURMULOTDIR/timidity/man.html"
  "HTML manual for Timidity"
  :type 'file
  :group 'spfa-paths)

(defun surmulot-timidity-man ()
  (expand-file-name (substitute-in-file-name surmulot-timidity-man)))

(defcustom surmulot-keykit-man "$SURMULOTDIR/keykit/index.html"
  "HTML manual for Keykit"
  :type 'file
  :group 'spfa-paths)

(defun surmulot-keykit-man ()
  (expand-file-name (substitute-in-file-name surmulot-keykit-man)))



;;; --------------------------- End  -----------------------------

(provide 'surmulot)

;;; surmulot.el ends here







